
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#define NPWM  200
#define PPWM  (160000 / NPWM)

uint8_t f50,odis;
uint16_t pcount,uf;
uint32_t vpwr,sst;
uint16_t l[NPWM+1];
float pwr,ch2,ch0,bq_ch0;
uint16_t dbounce;

double Input, Output, Setpoint;
double errSum, lastErr,error,dErr;
double kp, ki, kd;

float aa[6],d1[6],d2[6],w0[6],w1[6],w2[6],xv[10],yv[10];
int np;


void setup()
  {
  int i;
  float t,u;
  
  for(t=0.0,i=0; i <= NPWM; i++,t +=  3.14159/ (float) NPWM)
    {
    u = 16384.0 * sin(t);  
    l[i] = (int)u;  
    } 
    
  pinMode(9, OUTPUT);
  pinMode(10, OUTPUT);
  pinMode(8,INPUT_PULLUP);
  DDRD = 0xff;   // can NOT let arduino system alter any pins from D0 to D7 via calls to pinMode() etc.. It buggers d5 which I need to pull down low-Z upon boot
  noInterrupts();
  TCCR2A = 0; // I had some bits set which interfered with simple digital in PORTD. set to zero means no i/o pins driven by timer2.
  TCCR2B =   _BV(CS22)| _BV(CS21) | _BV(CS20);
  TIMSK2 |= (1 << TOIE2);
  TCCR1A = _BV(COM1A1)  | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 10;
  ICR1 = PPWM;
  TIMSK1 |= (1 << TOIE1);

  sst=0;
  f50 = 0;
  pcount = 0;
  uf=0;
  ch2=ch0=0.0;
  pwr=0.0;
  vpwr=0;
  dbounce = 0;
    // set up PID 0.1 0.0 0.01 for biquad LP Fc = 0.02, np=4
    kp = 0.1;   
    ki = 0.0;
    kd = 0.01;
    bq_ch0 = 0.0;
    
  errSum = 0.0;
  
  interrupts();             // enable all interrupts
  TIMSK0 = 0;                 // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;  

  Serial.begin(115200);
  }


ISR(TIMER2_OVF_vect)
  {// 100 Hz
  TCNT2=100;
  pcount=0;
  f50 = !f50;
  TCNT1=0;
  uf=1;
  if(odis == 1)
      sbi(PORTD,5);
    else
      cbi(PORTD,5);
  check_switch();  
  }

ISR(TIMER1_OVF_vect)        
  {//20Khz SPWM
  long c;  
  if(pcount == 1)
    {
    if (f50 == 1)
      {
      cbi(PORTB,2);
      sbi(PORTD,7);  // give a sync pulse for DSO, alligned with start of 50Hz
      cbi(PORTD,7);
      }
    else
      sbi(PORTB,2);
    } 
  c = (l[pcount] * vpwr) >> 14;   
  if (f50 == 1)
    OCR1A = c;
  else
    OCR1A = PPWM - c;
 pcount++;
}


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}

  
void loop()
{
  float s,fch0;
  float tc2=0.01; 
  uint16_t i,us;
  uint8_t b,c;
  // filter ch2,0 analog input and scale to 0 to 1
  // ch0 = output ACV, via transformer, 220VAC = 3.5V DC
  // ch2 = setpoint. About 4.3kHz sample rate, or 230us per 2ch sample and filter plus 2 x output to SPI DAC
  //ch2 = 0.57;  
  ch2 = (1.0-tc2)*ch2 + tc2 * analogRead(2)/1024.0;
     
  fch0 = analogRead(0)/1024.0;   
  // this is good, bessel filter of some Fc.
  xv[0] = xv[1]; xv[1] = xv[2]; xv[2] = xv[3]; xv[3] = xv[4]; 
  xv[4] = fch0 / 2.259747797e+05;
  yv[0] = yv[1]; yv[1] = yv[2]; yv[2] = yv[3]; yv[3] = yv[4]; 
  yv[4] =   (xv[0] + xv[4]) + 4 * (xv[1] + xv[3]) + 6 * xv[2]
                     + ( -0.7428578687 * yv[0]) + (  3.1954036268 * yv[1])
                     + ( -5.1599230905 * yv[2]) + (  3.7073065280 * yv[3]);
  bq_ch0 = yv[4];
        
  if(uf == 1 && odis == 1 )    // this will execute at 100Hz, thanks to uf = 1, set within 100Hz ISR(TIMER2_OVF_vect)
      { 
      ch0 = bq_ch0; 
      uf=0;
      if(sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {
          sst++;        // comment this line to prevent PID control, it will remain in simple bang-bang control
          if(ch0 < ch2)
          pwr += 1.0/300.0;
        else
          pwr -= 1.0/300.0;
        }
      else
        {
          Setpoint = ch2;
          Input = ch0;  
          do_pid();
          pwr = pwr + Output;
        }
      if (pwr > 1.0) pwr = 1.0;
      if (pwr < 0.01) pwr = 0.01;
      s = (float)PPWM * pwr;    
      vpwr = (int)s;  
      
      }
}




void check_switch()
{
dbounce++;  
if (dbounce > 100) dbounce = 100;
if (digitalRead(8) == 0)     // momentary button pressed, pulling internal pullup low
    {
    if (dbounce > 80)
      {  
      if(odis == 0) 
        {   // stop it.
        odis = 1;
        pwr = 0.0;
        vpwr = 0;
        sst=0;
        }
      else 
        {   // enable it. STAND BACK!
        odis = 0;
        pwr = 0.0;
        vpwr = 0;
        errSum=0.0;
        sst=0;
        }
     dbounce=0;
     }       
    }
}


